using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;

namespace WhackAMole
{
	/// <summary>
	/// Modal dialog box for creating the mapping between mouse IDs and mole IDs.
	/// </summary>
	public class CalibrationDialog : System.Windows.Forms.Form
	{
		private System.Windows.Forms.Label calibrationText;
		private Sdgt.SdgManager sdgManager;
		private System.ComponentModel.IContainer components;

		/// <summary>
		/// A constant for the mouse-to-mole map that specifies that a mouse has not
		/// been calibrated.
		/// </summary>
		private const int UNCALIBRATED = -50;

		/// <summary>
		/// A constant for the mouse-to-mole map that specifies the system mouse.
		/// </summary>
		private const int SYSTEMMOUSE = -1;

		/// <summary>
		/// The number of mice attached to the system.
		/// </summary>
		private int totalMice;

		/// <summary>
		/// The number of mice that have been calibrated.
		/// </summary>
		private int miceCalibrated;

		/// <summary>
		/// The mapping between mouse IDs and mole IDs.
		/// </summary>
		private int[] mapping;

		/// <summary>
		/// The ID of the system mouse.
		/// </summary>
		private int systemMouse;

		/// <summary>
		/// CalibrationDialog constructor.
		/// </summary>
		public CalibrationDialog()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			// determine how many mice are attached to the system
			totalMice = sdgManager.Mice.Count;

			// haven't calibrated any mice yet
			miceCalibrated = 0;

			// create space for mouse-to-mole map
			mapping = new int[totalMice];
			for(int i = 0; i < totalMice; i++)
			{
				mapping[i] = UNCALIBRATED;
			}
			calibrationText.Text = "You have " + totalMice + ((totalMice == 1) ? " mouse" : " mice") + " left to calibrate.";
			calibrationText.Text += "\n\nPlease click a button on your system mouse.  (Press ESC to cancel.)";
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.sdgManager = new Sdgt.SdgManager(this.components);
			this.calibrationText = new System.Windows.Forms.Label();
			this.SuspendLayout();
			// 
			// sdgManager
			// 
			this.sdgManager.EmulateSystemMouseMode = Sdgt.EmulateSystemMouseModes.Park;
			this.sdgManager.Keyboards = null;
			this.sdgManager.Mice = null;
			this.sdgManager.MouseToFollow = 0;
			this.sdgManager.ParkSystemMouseLocation = new System.Drawing.Point(10, 10);
			this.sdgManager.RelativeTo = this;
			this.sdgManager.MouseDown += new Sdgt.SdgMouseEventHandler(this.sdgManager_MouseDown);
			this.sdgManager.KeyDown += new Sdgt.SdgKeyEventHandler(this.sdgManager_KeyDown);
			// 
			// calibrationText
			// 
			this.calibrationText.Location = new System.Drawing.Point(16, 16);
			this.calibrationText.Name = "calibrationText";
			this.calibrationText.Size = new System.Drawing.Size(488, 112);
			this.calibrationText.TabIndex = 0;
			this.calibrationText.Text = "Please click a button on your system mouse.  (Press ESC to cancel.)";
			// 
			// CalibrationDialog
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(514, 144);
			this.Controls.Add(this.calibrationText);
			this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedDialog;
			this.Name = "CalibrationDialog";
			this.Text = "Calibrate";
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// Determines which mouse was clicked and which mole it should map to.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void sdgManager_MouseDown(object sender, Sdgt.SdgMouseEventArgs e)
		{
			// check if we've already calibrated the mouse that was clicked
			if(mapping[e.ID] != UNCALIBRATED)
			{
				calibrationText.Text = "You already calibrated that mouse.  Try again.";
			}
			else
			{
				// set up the mapping
				if(miceCalibrated == 0)
				{
					// set the system mouse to reference no mole
					mapping[e.ID] = SYSTEMMOUSE;
					systemMouse = e.ID;
				}
				else
				{
					// set the other mice to reference the appropriate mole in order
					mapping[e.ID] = miceCalibrated - 1;
				}

				miceCalibrated++;

				// display the number of remaining mice to the user
				int miceLeft = totalMice - miceCalibrated;
				calibrationText.Text = "You have " + miceLeft + ((miceLeft == 1) ? " mouse" : " mice") + " left to calibrate.";

				if(miceLeft == 0)
				{
					// all the mice have been calibrated, return an OK result
					calibrationText.Text += "\n\nCalibration complete!";
					sdgManager.Dispose();
					DialogResult = DialogResult.OK;
				}
				else if(miceCalibrated == 1)
				{
					calibrationText.Text += "\n\nPlease hit the leftmost mole with the bat.  (Press ESC to cancel.)";
				}
				else if(miceLeft > 0)
				{
					calibrationText.Text += "\n\nPlease hit the next mole to the right with the bat.  (Press ESC to cancel.)";
				}
			}
		}

		/// <summary>
		/// Handles keypresses in the dialog box.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void sdgManager_KeyDown(object sender, Sdgt.SdgKeyEventArgs e)
		{
			if(e.KeyCode == Keys.Escape)
			{
				sdgManager.Dispose();
				DialogResult = DialogResult.Cancel;
			}
		}

		/// <summary>
		/// Gets the mapping between mouse IDs and mole IDs.
		/// </summary>
		/// <value>
		/// Mouse-to-mole map.
		/// </value>
		public object Mapping
		{
			get
			{
				return mapping;
			}
		}

		/// <summary>
		/// Gets the system mouse ID.
		/// </summary>
		/// <value>
		/// The system mouse ID.
		/// </value>
		public int SystemMouse
		{
			get
			{
				return systemMouse;
			}
		}

		public new DialogResult ShowDialog()
		{
			// check if there are enough mice attached to the system (makes no sense to play if there
			// is only one system mouse and there are no moles)
			if(totalMice < 2)
			{
				sdgManager.Dispose();
				MessageBox.Show("You need to have at least 2 mice connected to your system to play.");
				return DialogResult.Cancel;
			}

			return base.ShowDialog();
		}
	}
}