using System;
using System.Diagnostics;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Threading;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectSound;

namespace WhackAMole
{
	/// <summary>
	/// Specifies state for the game's state machine.
	/// </summary>
	enum GameState { NotStarted, TrainingMode, GameMode, GameOver };
	
	/// <summary>
	/// The form for the WhackAMole game.
	/// </summary>
	public class GameForm : System.Windows.Forms.Form
	{
		private System.ComponentModel.IContainer components;

		/// <summary>
		/// An array of moles.
		/// </summary>
		private Mole[] moles;

		/// <summary>
		/// The calibration information provided by the user.
		/// </summary>
		private CalibrationInfo calibrationInfo;

		/// <summary>
		/// The number of moles currently up.
		/// </summary>
		private int numMolesUp = 0;

		/// <summary>
		/// The number of moles that have been hit in the current wave of moles.
		/// </summary>
		private int molesHit = 0;

		/// <summary>
		/// A random number generator.
		/// </summary>
		private static Random randGen = new Random();

		/// <summary>
		/// The SDG Toolkit manager.
		/// </summary>
		private Sdgt.SdgManager sdgManager;

		private System.Windows.Forms.Label pressEscLabel;

		/// <summary>
		/// The current game state.
		/// </summary>
		private GameState currentState;

		/// <summary>
		/// The game loop thread.
		/// </summary>
		private Thread gameThread = null;

		/// <summary>
		/// Notifies the game thread that an event has occurred.  Used mostly to determine
		/// when sounds have finished playing.
		/// </summary>
		private AutoResetEvent arey = new AutoResetEvent(false);

		/// <summary>
		/// A simple method to shuffle the entries of an array.
		/// </summary>
		/// <param name="arr">the array to shuffle</param>
		private static void Shuffle(object[] arr)
		{
			for(int i = 0; i < arr.Length; i++)
			{
				int r = randGen.Next(0, arr.Length);
				object temp = arr[i];
				arr[i] = arr[r];
				arr[r] = temp;
			}
		}

		/// <summary>
		/// Waits for the specified sound buffer to complete playing.
		/// </summary>
		/// <param name="buffer">the sound buffer to wait for</param>
		private void WaitForEndOfSound(SecondaryBuffer buffer)
		{
			while(buffer.Status.Playing)
			{
				// let other threads run
				Thread.Sleep(0);
			}
		}

		/// <summary>
		/// The main game loop.
		/// </summary>
		public void GameLoop()
		{
			try
			{
				// keep going until the thread has been interrupted (i.e., the user
				// has ended the game by closing the window)
				while(true)
				{
					// sound buffer used for instructions
					SecondaryBuffer buffer = null;

					// keep looping intro and training information until the user requests to
					// start a game
					while(currentState != GameState.GameMode)
					{
						// haven't started either training or game yet
						currentState = GameState.NotStarted;

						// Ask if they know how to play.  If they do, they should hit a button within
						// 5 seconds to start the game.
						// Otherwise go into training mode.

						// Play sound: "Welcome to Whack-a-Mole! If you already know how to play,
						// hit a mole now to start playing!  If you want to learn how to play, wait
						// and we'll show you how!"
						buffer = new SecondaryBuffer(@"../sounds/WelcomeToWhackAMole.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);

						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						// Wait for up to 5 seconds for a mole to be hit.
						arey.WaitOne(5000, true);
						if(currentState == GameState.GameMode)
						{
							// the user hit a mole, requesting to start a game, so break out of the
							// training loop
							break;
						}

						// start training mode
						currentState = GameState.TrainingMode;

						// Play sound: "Mama, Papa, Billy, and Milly Mole are tearing up your garden!
						// You need to stop them from doing that.  Here's how."
						buffer = new SecondaryBuffer(@"../sounds/TrainingTearingUpGarden.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						// Play sound: "When you hear this sound,"
						buffer = new SecondaryBuffer(@"../sounds/WhenYouHear.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						// Play sound: "Get Up Sound"
						buffer = new SecondaryBuffer(@"../sounds/gup.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						// Play sound: "it means that a mole has popped up in your garden!  After they pop up,
						// they will start to bother you."
						buffer = new SecondaryBuffer(@"../sounds/TrainingWheneverYouHearGetUp.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						// Play sound: "Mama mole is here!  Can you hear that she is to your left?  Reach out
						// with your bat and see if you can hit her!"
						buffer = new SecondaryBuffer(@"../sounds/TrainingMamaaIsHere.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						// HACK: This is hardcoded for 4 moles!
						// TODO: also should have a sound to indicate when the user missed!
						moles[0].GetUp();
						Thread.Sleep(500);
						moles[0].Taunt();
						arey.WaitOne();

						// Play sound: "Good job!  Papa mole is here!  Reach out with your bat and see if you can hit him!"
						buffer = new SecondaryBuffer(@"../sounds/TrainingPapaIsHere.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						moles[1].GetUp();
						Thread.Sleep(500);
						moles[1].Taunt();
						arey.WaitOne();

						// Play sound: "Good job!  Billy mole is here!  Reach out with your bat and see if you can hit him!"
						buffer = new SecondaryBuffer(@"../sounds/TrainingBillyIsHere.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						moles[2].GetUp();
						Thread.Sleep(500);
						moles[2].Taunt();
						arey.WaitOne();

						// Play sound: "Good job!  Milly mole is here!  Reach out with your bat and see if you can hit him!"
						buffer = new SecondaryBuffer(@"../sounds/TrainingMillyIsHere.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);

						moles[3].GetUp();
						Thread.Sleep(500);
						moles[3].Taunt();
						arey.WaitOne();

						// Play sound: "And that's all there is to it!  Keep getting those moles, and you'll go
						// to higher levels!"
						buffer = new SecondaryBuffer(@"../sounds/TrainingAndThatIsAll.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
						buffer.Play(0, BufferPlayFlags.Default);
						// Wait for end of sound.
						WaitForEndOfSound(buffer);
					}

					// the actual game starts here

					// HACK: Hardcoded for 4 levels!

					// waitTime is used to specify how long a mole should be up before hiding again
					int waitTime = 0;
					for(int level = 0; level < 4; level++)
					{
						molesHit = 0;

						// play the level sound
						switch(level)
						{
							case 0:
								waitTime = 2000;
								buffer = new SecondaryBuffer(@"../sounds/level1.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
								break;
							case 1:
								waitTime = 1500;
								buffer = new SecondaryBuffer(@"../sounds/level2.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
								break;
							case 2:
								waitTime = 1000;
								buffer = new SecondaryBuffer(@"../sounds/level3.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
								break;
							case 3:
								waitTime = 750;
								buffer = new SecondaryBuffer(@"../sounds/level4.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
								break;
						}

						buffer.Play(0, BufferPlayFlags.Default);

						// wait for sound to end
						WaitForEndOfSound(buffer);

						// start waves of moles, bringing up X amount of moles depending on level
						// enable mole timer
						// currently hardcoded for 1 mole at a time, and 10 moles per level
						for(int moleWave = 0; moleWave < 10; moleWave++)
						{
							// should be 2 moles up for levels 3 and 4, need to figure out how to do this
							numMolesUp = 1;

							// pick a random mole to bring up
							int r = randGen.Next(0, calibrationInfo.NumMoles);

							Trace.WriteLine("r = " + r);

							// make the mole get up
							moles[r].GetUp();
							Thread.Sleep(500);

							// taunt the player
							moles[r].Taunt();

							// wait for up to the waiting time for the user to hit the mole
							arey.WaitOne(waitTime, true);

							// hide the mole (only plays a sound if the mole hasn't been hit)
							moles[r].Hide();
							Thread.Sleep(500);

							// shuffling only needed if I make the papa worth more points
							// - left it in for now
							Shuffle(moles);

							// the moles have been shuffled, so need to reset their pan positions
							// appropriately
							for(int i = 0; i < calibrationInfo.NumMoles; i++)
							{
								moles[i].PanPosition = calibrationInfo.PanPositions[i];
							}
						}
						// based on how many moles they hit properly (and/or how many they missed), determine
						// if they should move on to the next level.
						// if they shouldn't, play "game over"-ish sound and break out of the for loop?
						// if they should, play "good job"? and continue on to next level?

						bool gameOver = false;

						// these could be done with gameOver = (molesHit < 7); lines, but
						// I left it this way for now
						switch(level)
						{
							case 0:
								if(molesHit < 7)
								{
									// game over
									gameOver = true;
								}
								else
								{
									gameOver = false;
								}
								break;
							case 1:
								if(molesHit < 6)
								{
									// game over
									gameOver = true;
								}
								else
								{
									gameOver = false;
								}
								break;
							case 2:
								if(molesHit < 5)
								{
									// game over
									gameOver = true;
								}
								else
								{
									gameOver = false;
								}
								break;
							case 3:
								if(molesHit < 5)
								{
									// game over
									gameOver = true;
								}
								else
								{
									gameOver = false;
								}
								break;
						}

						if(gameOver)
						{
							currentState = GameState.GameOver;
							// Play sound: "Oh no!  Too many moles got away that time.  To try again,
							// whack a mole!"
							buffer = new SecondaryBuffer(@"../sounds/OhNoTryAgain.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
							buffer.Play(0, BufferPlayFlags.Default);

							// Wait for end of sound.
							WaitForEndOfSound(buffer);

							// wait for up to 5 seconds for the user to hit a mole to try again
							arey.WaitOne(5000, true);

							if(currentState == GameState.GameMode)
							{
								// the user hit a mole, start the game over
								level = -1;
								continue;
							}

							// break out of the loop and go back to training mode
							break;
						}
						else
						{
							// the game isn't over yet
							if(level == 3)
							{
								// the player passed the last level
								currentState = GameState.GameOver;

								// Play sound: "Congratulations!  You scared all of the moles out of
								// your garden."
								buffer = new SecondaryBuffer(@"../sounds/WinDaGame.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
								buffer.Play(0, BufferPlayFlags.Default);

								// Wait for end of sound.
								WaitForEndOfSound(buffer);
							}
							else
							{
								// Play sound: "Congratulations!  You made it to the next level!"
								buffer = new SecondaryBuffer(@"../sounds/MadeItToNextLevel.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
								buffer.Play(0, BufferPlayFlags.Default);

								// Wait for end of sound.
								WaitForEndOfSound(buffer);
							}
						}
					}

					// play game over sound/tell them what score they got, what level they achieved, etc.
				}
			}
			catch(ThreadAbortException)
			{
				// the user has requested that the game end, hide all the moles
				for(int i = 0; i < calibrationInfo.NumMoles; i++)
				{
					moles[i].Hide();
				}
			}
		}

		/// <summary>
		/// GameForm constructor.
		/// </summary>
		/// <param name="calibrationInfo">
		/// A CalibrationInfo instance that contains mouse and speaker calibration information.
		/// </param>
		public GameForm(CalibrationInfo calibrationInfo)
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			// set up the supermouse
//			sdgManager.MouseToFollow = 3;

			// HACK: there seems to be a bug in SDGT that doesn't set up hotspots properly.
			// make sure the hotspots for each mouse cursor are at the proper place (upper left mouse point)
			foreach(Sdgt.Mouse mouse in sdgManager.Mice)
			{
				mouse.Cursor = Cursors.Default;
				mouse.Hotspot = new Point(10, 11);
				mouse.Visible = false;
			}

			this.calibrationInfo = calibrationInfo;

			// create the moles
			moles = new Mole[calibrationInfo.NumMoles];
			for(int i = 0; i < calibrationInfo.NumMoles; i++)
			{
				// associate sounds with each mole (each one has to have its own instances so that
				// they can be played simultaneously)
				SecondaryBuffer getUpBuf = new SecondaryBuffer(@"../sounds/gup.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
				SecondaryBuffer hideBuf = new SecondaryBuffer(@"../sounds/gdown.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
				SecondaryBuffer hitBuf = new SecondaryBuffer(@"../sounds/ow.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);
				SecondaryBuffer tauntBuf = new SecondaryBuffer(@"../sounds/laugh.wav", calibrationInfo.BufferDesc, calibrationInfo.SoundDevice);

				moles[i] = new Mole();
				moles[i].SoundGetUp = getUpBuf;
				moles[i].SoundHide = hideBuf;
				moles[i].SoundHit = hitBuf;
				moles[i].SoundTaunt = tauntBuf;

				// set up the pan positions from the calibration information
				moles[i].PanPosition = calibrationInfo.PanPositions[i];
			}

			// set up the mole names (is this necessary?)
			for(int i = 0; i < calibrationInfo.NumMoles; i++)
			{
				// TODO: hardcoded for 4 moles!  bad!
				switch(i)
				{
					case 0:
						moles[i].Name = "Mama";
						break;
					case 1:
						moles[i].Name = "Papa";
						break;
					case 2:
						moles[i].Name = "Billy";
						break;
					case 3:
						moles[i].Name = "Milly";
						break;
					default:
						moles[i].Name = "Acme";
						break;
				}
			}

			// put the game into the initial NotStarted state
			currentState = GameState.NotStarted;

			// start the main game thread loop
			gameThread = new Thread(new ThreadStart(GameLoop));
			gameThread.Start();
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.sdgManager = new Sdgt.SdgManager(this.components);
			this.pressEscLabel = new System.Windows.Forms.Label();
			this.SuspendLayout();
			// 
			// sdgManager
			// 
			this.sdgManager.EmulateSystemMouseMode = Sdgt.EmulateSystemMouseModes.Park;
			this.sdgManager.Keyboards = null;
			this.sdgManager.Mice = null;
			this.sdgManager.MouseToFollow = 0;
			this.sdgManager.ParkSystemMouseLocation = new System.Drawing.Point(10, 10);
			this.sdgManager.RelativeTo = this;
			this.sdgManager.MouseDown += new Sdgt.SdgMouseEventHandler(this.sdgManager_MouseDown);
			this.sdgManager.KeyDown += new Sdgt.SdgKeyEventHandler(this.sdgManager_KeyDown);
			// 
			// pressEscLabel
			// 
			this.pressEscLabel.Location = new System.Drawing.Point(8, 8);
			this.pressEscLabel.Name = "pressEscLabel";
			this.pressEscLabel.Size = new System.Drawing.Size(496, 24);
			this.pressEscLabel.TabIndex = 4;
			this.pressEscLabel.Text = "Press ESC to quit the game.";
			// 
			// GameForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(512, 86);
			this.Controls.Add(this.pressEscLabel);
			this.Name = "GameForm";
			this.Text = "GameForm";
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The mouse button click handler.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void sdgManager_MouseDown(object sender, Sdgt.SdgMouseEventArgs e)
		{
			// check a mouse button from a non-system mouse was clicked
			if(e.ID != calibrationInfo.MouseToMoleMap.Length - 1)
			{
				// non-system mouse clicked

				if(currentState == GameState.NotStarted)
				{
					// we haven't started the game yet, and the user hit a mole,
					// so start the game
					currentState = GameState.GameMode;
					arey.Set();
				}
				else if(currentState == GameState.GameMode)
				{
					// we are in the middle of a game, so check which mole was hit
					if(moles[calibrationInfo.MouseToMoleMap[e.ID]].IsUp)
					{
						// the mole that was hit was up, so hit it
						molesHit++;
						moles[calibrationInfo.MouseToMoleMap[e.ID]].Hit();
						numMolesUp--;
						if(numMolesUp == 0)
						{
							// the game loop thread is waiting for all of the moles to be hit
							arey.Set();
						}
					}
				}
				else if(currentState == GameState.TrainingMode)
				{
					// we are in training mode, so check which mole was hit for training purposes
					if(moles[calibrationInfo.MouseToMoleMap[e.ID]].IsUp)
					{
						// the mole that was hit was up, so hit it
						moles[calibrationInfo.MouseToMoleMap[e.ID]].Hit();
						arey.Set();
					}
				}
				else if(currentState == GameState.GameOver)
				{
					// the game is over, and the user hit a mole to restart the game
					currentState = GameState.GameMode;
					arey.Set();
				}
			}
		}

		/// <summary>
		/// Keypress handler.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void sdgManager_KeyDown(object sender, Sdgt.SdgKeyEventArgs e)
		{
			// if the user presses the ESC key, abort the game thread and wait for it
			// to finish, and then close the form
			if(e.KeyCode == Keys.Escape) 
			{
				Trace.WriteLine("ESC");
				gameThread.Abort();
				gameThread.Join();
				this.Close();
			}
		}
	}
}