using System;
using System.Diagnostics;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectSound;

namespace WhackAMole
{
	/// <summary>
	/// The main Whack-a-Mole form.
	/// </summary>
	public class MainForm : System.Windows.Forms.Form
	{
		private System.ComponentModel.IContainer components = null;

		/// <summary>
		/// An instance of GameForm.
		/// </summary>
		private GameForm gameForm = null;

		/// <summary>
		/// A device handle for sound output.
		/// </summary>
		private Device device;

		/// <summary>
		/// A buffer description used to control sound buffer output properties.
		/// </summary>
		private BufferDescription desc;

		/// <summary>
		/// Contains the mapping between mouse ID and mole ID.
		/// </summary>
		private int[] mouseToMoleMap;

		/// <summary>
		/// The ID of the system mouse.
		/// </summary>
		private int systemMouse;

		private System.Windows.Forms.Button buttonCalibrate;
		private System.Windows.Forms.Button buttonPlay;
		private System.Windows.Forms.Button buttonQuit;

		/// <summary>
		/// Indicates whether or not the mice and speakers have been calibrated.
		/// </summary>
		private bool calibrated = false;

		/// <summary>
		/// Stores mouse and speaker calibration information.
		/// </summary>
		private CalibrationInfo calibrationInfo;

		/// <summary>
		/// MainForm constructor.
		/// </summary>
		public MainForm()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			// don't want to turn on the DebugConsole in this version
/*#if (DEBUG)
			// debug mode
			DebugConsole.Instance.Init(true,true);
#else
			// release mode
			DebugConsole.Instance.Init(false,true);
#endif
*/
			Trace.WriteLine("--- LOG START ---");

			// set up the sound device
			device = new Device();
			device.SetCooperativeLevel(this, CooperativeLevel.Normal);

			// set up the buffer description to allow for panning control as well as
			// background playback
			desc = new BufferDescription();
			desc.ControlPan = true;
			desc.GlobalFocus = true;
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.buttonCalibrate = new System.Windows.Forms.Button();
			this.buttonPlay = new System.Windows.Forms.Button();
			this.buttonQuit = new System.Windows.Forms.Button();
			this.SuspendLayout();
			// 
			// buttonCalibrate
			// 
			this.buttonCalibrate.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.buttonCalibrate.Location = new System.Drawing.Point(8, 8);
			this.buttonCalibrate.Name = "buttonCalibrate";
			this.buttonCalibrate.Size = new System.Drawing.Size(152, 56);
			this.buttonCalibrate.TabIndex = 0;
			this.buttonCalibrate.Text = "Game Setup";
			this.buttonCalibrate.Click += new System.EventHandler(this.buttonCalibrate_Click);
			// 
			// buttonPlay
			// 
			this.buttonPlay.Enabled = false;
			this.buttonPlay.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.buttonPlay.Location = new System.Drawing.Point(176, 8);
			this.buttonPlay.Name = "buttonPlay";
			this.buttonPlay.Size = new System.Drawing.Size(152, 56);
			this.buttonPlay.TabIndex = 1;
			this.buttonPlay.Text = "Play!";
			this.buttonPlay.Click += new System.EventHandler(this.buttonPlay_Click);
			// 
			// buttonQuit
			// 
			this.buttonQuit.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.buttonQuit.Location = new System.Drawing.Point(344, 8);
			this.buttonQuit.Name = "buttonQuit";
			this.buttonQuit.Size = new System.Drawing.Size(152, 56);
			this.buttonQuit.TabIndex = 2;
			this.buttonQuit.Text = "Quit";
			this.buttonQuit.Click += new System.EventHandler(this.buttonQuit_Click);
			// 
			// MainForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(504, 78);
			this.Controls.Add(this.buttonQuit);
			this.Controls.Add(this.buttonPlay);
			this.Controls.Add(this.buttonCalibrate);
			this.Name = "MainForm";
			this.RightToLeft = System.Windows.Forms.RightToLeft.No;
			this.Text = "Whack-a-Mole!";
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		public static void Main() 
		{
			try
			{
				// these two lines are needed to make the form buttons look pretty
				Application.EnableVisualStyles();
				Application.DoEvents();

				// start up the main form
				Application.Run(new MainForm());
			}
			catch(Exception e)
			{
				MessageBox.Show(e.ToString());
			}
		}

		/// <summary>
		/// Calibrates the mice and speakers attached to the system.
		/// </summary>
		private void calibrateMice()
		{
			// create a new CalibrationDialog to create the mouse-to-mole map
			CalibrationDialog cd = new CalibrationDialog();

			if(cd.ShowDialog() == DialogResult.OK)
			{
				// the user successfully calibrated the mice, to set up the mapping
				mouseToMoleMap = (int[]) cd.Mapping;
				systemMouse = cd.SystemMouse;

				// proceed to speaker calibration
				SpeakerCalibrationDialog scd = new SpeakerCalibrationDialog(mouseToMoleMap.Length - 1, device, desc);
				scd.ShowDialog();

				// create a new CalibrationInfo object with the mouse and speaker calibration
				// information to pass into the game
				calibrationInfo = new CalibrationInfo(mouseToMoleMap.Length - 1, mouseToMoleMap, scd.PanPositions, device, desc);

				// debugging output
				for(int i = 0; i < mouseToMoleMap.Length; i++)
				{
					Trace.WriteLine("[mouse " + i + "]: mole " + mouseToMoleMap[i]);
				}
				Trace.WriteLine("System mouse: " + systemMouse);

				for(int i = 0; i < mouseToMoleMap.Length - 1; i++)
				{
					Trace.WriteLine("pan[" + i + "]: " + scd.PanPositions[i]);
				}

				// now that we're calibrated, enable the Play! button
				buttonPlay.Enabled = true;
				calibrated = true;
			}
			// make this form request the focus after the dialog boxes close --
			// for some reason focus is not returned properly when using SDGT
			Focus();
		}

		/// <summary>
		/// Button handler for "Calibrate" button.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void buttonCalibrate_Click(object sender, System.EventArgs e)
		{
			calibrateMice();
		}

		/// <summary>
		/// Button handler for "Play!" button.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void buttonPlay_Click(object sender, System.EventArgs e)
		{
			// don't let the user play without first calibrating
			if(!calibrated)
			{
				calibrateMice();
			}

			// now check if the user has successfully calibrated
			if(calibrated)
			{
				// everything's good, start up the game
				gameForm = new GameForm(calibrationInfo);
				gameForm.Show();
			}
		}

		/// <summary>
		/// Button handler for "Quit" button.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void buttonQuit_Click(object sender, System.EventArgs e)
		{
			// close this form
			Close();
		}
	}
}
